"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.constructClaimTransaction = void 0;
const bitcoin_ops_1 = __importDefault(require("@boltz/bitcoin-ops"));
const bip371_1 = require("bitcoinjs-lib/src/psbt/bip371");
const liquidjs_lib_1 = require("liquidjs-lib");
const bufferutils_1 = require("liquidjs-lib/src/bufferutils");
const Utils_1 = require("../../Utils");
const Enums_1 = require("../../consts/Enums");
const Claim_1 = require("../../swap/Claim");
const SwapUtils_1 = require("../../swap/SwapUtils");
const Utils_2 = require("../Utils");
const Networks_1 = __importDefault(require("../consts/Networks"));
const init_1 = require("../init");
const TaprootUtils_1 = require("./TaprootUtils");
const dummyTaprootSignature = Buffer.alloc(64);
const getSighashType = (type) => type === Enums_1.OutputType.Taproot
    ? liquidjs_lib_1.Transaction.SIGHASH_DEFAULT
    : liquidjs_lib_1.Transaction.SIGHASH_ALL;
const validateLiquidInputs = (utxos, isRefund) => {
    (0, Claim_1.validateInputs)(utxos);
    const taprootInputs = utxos.filter(Claim_1.isRelevantTaprootOutput);
    if (isRefund && taprootInputs.some((utxo) => utxo.keys === undefined)) {
        throw 'not all Taproot refund inputs have keys';
    }
    if (taprootInputs.some((utxo) => utxo.keys === undefined &&
        utxo.swapTree.covenantClaimLeaf === undefined)) {
        throw 'not all Taproot signature claims have keys';
    }
};
/**
 * Claim swaps
 *
 * @param utxos UTXOs that should be claimed or refunded
 * @param destinationScript the output script to which the funds should be sent
 * @param fee how many satoshis should be paid as fee
 * @param isRbf whether the transaction should signal full Replace-by-Fee
 * @param network network of the transaction
 * @param blindingKey blinding public key for the output; undefined if the output should not be blinded
 * @param timeoutBlockHeight locktime of the transaction; only needed if the transaction is a refund
 * @param isRefund whether the transaction is a refund or claim
 */
const constructClaimTransaction = (utxos, destinationScript, fee, isRbf = true, network = Networks_1.default.liquidMainnet, blindingKey, timeoutBlockHeight, isRefund = false) => {
    validateLiquidInputs(utxos, isRefund);
    if (utxos.some((utxo) => (utxo.blindingPrivateKey === undefined) !==
        (utxos[0].blindingPrivateKey === undefined))) {
        throw 'all or none inputs have to be blinded';
    }
    if (utxos.some((utxo) => utxo.type !== Enums_1.OutputType.Taproot && utxo.type !== Enums_1.OutputType.Bech32)) {
        throw 'only Taproot or native SegWit inputs supported';
    }
    const pset = liquidjs_lib_1.Creator.newPset();
    const updater = new liquidjs_lib_1.Updater(pset);
    let utxoValueSum = BigInt(0);
    for (const [i, utxo] of utxos.entries()) {
        utxoValueSum += BigInt((0, Utils_2.getOutputValue)(utxo));
        const txHash = Buffer.alloc(utxo.txHash.length);
        utxo.txHash.copy(txHash);
        const txid = (0, Utils_1.getHexString)((0, bufferutils_1.reverseBuffer)(txHash));
        pset.addInput(new liquidjs_lib_1.CreatorInput(txid, utxo.vout, isRbf ? 0xfffffffd : 0xffffffff, i == 0 ? timeoutBlockHeight : undefined).toPartialInput());
        updater.addInSighashType(i, getSighashType(utxo.type));
        if (utxo.type === Enums_1.OutputType.Legacy) {
            updater.addInNonWitnessUtxo(i, utxo.legacyTx);
            updater.addInRedeemScript(i, utxo.redeemScript);
        }
        else if (utxo.type === Enums_1.OutputType.Compatibility) {
            updater.addInNonWitnessUtxo(i, utxo.legacyTx);
            updater.addInRedeemScript(i, (0, SwapUtils_1.scriptBuffersToScript)([
                (0, SwapUtils_1.scriptBuffersToScript)([
                    (0, Utils_1.getHexString)(bufferutils_1.varuint.encode(bitcoin_ops_1.default.OP_0)),
                    liquidjs_lib_1.crypto.sha256(utxo.redeemScript),
                ]),
            ]));
        }
        if (utxo.type !== Enums_1.OutputType.Legacy) {
            updater.addInWitnessUtxo(i, utxo);
            updater.addInWitnessScript(i, utxo.redeemScript);
        }
    }
    updater.addOutputs([
        {
            script: destinationScript,
            blindingPublicKey: blindingKey,
            asset: network.assetHash,
            amount: Number(utxoValueSum - BigInt(fee)),
            blinderIndex: blindingKey !== undefined ? 0 : undefined,
        },
    ]);
    const addFeeOutput = (isUnblinded = false) => {
        updater.addOutputs([
            {
                amount: isUnblinded ? fee - 1 : fee,
                asset: network.assetHash,
            },
        ]);
    };
    if (utxos[0].blindingPrivateKey !== undefined) {
        // We have to have at least one blinded output if we are spending blinded coins,
        // so we add a small OP_RETURN
        if (blindingKey === undefined) {
            pset.addOutput(new liquidjs_lib_1.CreatorOutput(network.assetHash, 
            // TODO: figure out flakiness with blinding 0 amount outputs
            1, Buffer.of(bitcoin_ops_1.default.OP_RETURN), Buffer.from(init_1.ecpair.makeRandom().publicKey), 0).toPartialOutput());
        }
        addFeeOutput(blindingKey === undefined);
    }
    else {
        addFeeOutput();
    }
    if (utxos[0].blindingPrivateKey !== undefined || blindingKey !== undefined) {
        blindPset(pset, utxos);
    }
    const signer = new liquidjs_lib_1.Signer(pset);
    const signatures = [];
    for (const [i, utxo] of utxos.entries()) {
        if (utxo.type === Enums_1.OutputType.Taproot) {
            if (utxo.cooperative || utxo.keys === undefined) {
                signatures.push(dummyTaprootSignature);
                continue;
            }
            const leafHash = (0, TaprootUtils_1.tapLeafHash)(isRefund ? utxo.swapTree.refundLeaf : utxo.swapTree.claimLeaf);
            const signature = utxo.keys.signSchnorr(pset.getInputPreimage(i, getSighashType(utxo.type), network.genesisBlockHash, leafHash));
            signatures.push(Buffer.from(signature));
            signer.addSignature(i, {
                genesisBlockHash: network.genesisBlockHash,
                tapScriptSigs: [
                    {
                        signature: Buffer.from(signature),
                        pubkey: (0, bip371_1.toXOnly)(Buffer.from(utxo.keys.publicKey)),
                        leafHash,
                    },
                ],
            }, liquidjs_lib_1.Pset.SchnorrSigValidator(init_1.secp.ecc));
        }
        else {
            const signature = liquidjs_lib_1.script.signature.encode(Buffer.from(utxo.keys.sign(Buffer.from(pset.getInputPreimage(i, getSighashType(utxo.type))))), getSighashType(utxo.type));
            signatures.push(signature);
            signer.addSignature(i, {
                partialSig: {
                    pubkey: Buffer.from(utxo.keys.publicKey),
                    signature,
                },
            }, liquidjs_lib_1.Pset.ECDSASigValidator(init_1.secp.ecc));
        }
    }
    const finalizer = new liquidjs_lib_1.Finalizer(pset);
    for (const [i, utxo] of utxos.entries()) {
        finalizer.finalizeInput(i, () => {
            const finals = {};
            if (utxo.type === Enums_1.OutputType.Legacy) {
                finals.finalScriptSig = (0, SwapUtils_1.scriptBuffersToScript)([
                    signatures[i],
                    utxo.preimage,
                    bitcoin_ops_1.default.OP_PUSHDATA1,
                    utxo.redeemScript,
                ]);
            }
            else if (utxo.type === Enums_1.OutputType.Compatibility) {
                finals.finalScriptSig = (0, SwapUtils_1.scriptBuffersToScript)([
                    (0, SwapUtils_1.scriptBuffersToScript)([
                        (0, Utils_1.getHexString)(bufferutils_1.varuint.encode(bitcoin_ops_1.default.OP_0)),
                        liquidjs_lib_1.crypto.sha256(utxo.redeemScript),
                    ]),
                ]);
            }
            if (utxo.type === Enums_1.OutputType.Taproot) {
                if (utxo.cooperative) {
                    // Add a dummy to allow for extraction and an accurate fee estimation
                    finals.finalScriptWitness = (0, liquidjs_lib_1.witnessStackToScriptWitness)([
                        dummyTaprootSignature,
                    ]);
                }
                else {
                    const isCovenantClaim = utxo.keys === undefined;
                    const tapleaf = isRefund
                        ? utxo.swapTree.refundLeaf
                        : isCovenantClaim
                            ? utxo.swapTree.covenantClaimLeaf
                            : utxo.swapTree.claimLeaf;
                    const witness = isRefund
                        ? [signatures[i]]
                        : isCovenantClaim
                            ? [utxo.preimage]
                            : [signatures[i], utxo.preimage];
                    finals.finalScriptWitness = (0, liquidjs_lib_1.witnessStackToScriptWitness)(witness.concat([
                        tapleaf.output,
                        (0, TaprootUtils_1.createControlBlock)((0, TaprootUtils_1.toHashTree)(utxo.swapTree.tree), tapleaf, utxo.internalKey),
                    ]));
                }
            }
            else if (utxo.type !== Enums_1.OutputType.Legacy) {
                finals.finalScriptWitness = (0, liquidjs_lib_1.witnessStackToScriptWitness)([
                    signatures[i],
                    utxo.preimage,
                    utxo.redeemScript,
                ]);
            }
            return finals;
        });
    }
    return liquidjs_lib_1.Extractor.extract(pset);
};
exports.constructClaimTransaction = constructClaimTransaction;
const blindPset = (pset, utxos) => {
    const zkpGenerator = new liquidjs_lib_1.ZKPGenerator(init_1.secp, liquidjs_lib_1.ZKPGenerator.WithBlindingKeysOfInputs(utxos.map((utxo) => utxo.blindingPrivateKey)));
    const zkpValidator = new liquidjs_lib_1.ZKPValidator(init_1.secp);
    const outputBlindingArgs = zkpGenerator.blindOutputs(pset, liquidjs_lib_1.Pset.ECCKeysGenerator(init_1.secp.ecc));
    const blinder = new liquidjs_lib_1.Blinder(pset, zkpGenerator.unblindInputs(pset), zkpValidator, zkpGenerator);
    blinder.blindLast({ outputBlindingArgs });
};
//# sourceMappingURL=Claim.js.map